/*
 * cmsfs-fuse - CMS EDF filesystem support for Linux
 * Main functions.
 *
 * Copyright IBM Corp. 2009
 * Author(s): Jan Glauber <jang@linux.vnet.ibm.com>
 */

#define FUSE_USE_VERSION 26
#define _GNU_SOURCE
#include <unistd.h>
#include <stdio.h>
#include <strings.h>
#include <string.h>
#include <stddef.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <time.h>
#include <assert.h>
#include <fcntl.h>
#include <search.h>
#include <fuse.h>
#include <fuse_opt.h>

#include "zt_common.h"
#include "cmsfs-fuse.h"
#include "ebcdic.h"
#include "filetypes.h"

struct cmsfs cmsfs;
FILE *logfile;

#define FSNAME_MAX_LEN	50

#define CMSFS_OPT(t, p, v) { t, offsetof(struct cmsfs, p), v }

enum {
	KEY_HELP,
	KEY_VERSION,
};

static const struct fuse_opt cmsfs_opts[] = {
	CMSFS_OPT("-a",			ascii_mode, 1),
	CMSFS_OPT("--ascii",		ascii_mode, 1),
	CMSFS_OPT("-t",			filetype_mode, 1),
	CMSFS_OPT("--filetype",		filetype_mode, 1),

	FUSE_OPT_KEY("-h",		KEY_HELP),
	FUSE_OPT_KEY("--help",		KEY_HELP),
	FUSE_OPT_KEY("-v",		KEY_VERSION),
	FUSE_OPT_KEY("--version",	KEY_VERSION),
	FUSE_OPT_END
};

static void usage(const char *progname)
{
	fprintf(stderr,
"Usage: %s DEVICE MOUNTPOINT [OPTIONS]\n"
"\n"
"Use the cmsfs-fuse command to read files stored on a z/VM CMS disk.\n"
"\n"
"General options:\n"
"    -o opt,[opt...]        Mount options\n"
"    -h   --help            Print help, then exit\n"
"    -v   --version         Print version, then exit\n"
"    -t   --filetype        Ascii translation based on file type\n"
"    -a   --ascii           Force ascii translation\n"
"\n", progname);
}

#define REAL(x)			((x - 1) * cmsfs.blksize)
#define UNREAL(x)		((x / cmsfs.blksize) + 1)

#define PTR_SIZE		(sizeof(struct fixed_ptr))
#define PTRS_PER_BLOCK		(cmsfs.fixed_ptrs_per_block)
#define VPTR_SIZE		(sizeof(struct var_ptr))
#define VPTRS_PER_BLOCK		(cmsfs.var_ptrs_per_block)
#define BITS_PER_DATA_BLOCK	(cmsfs.bits_per_data_block)
#define BITS_PER_PTR_BLOCK	(cmsfs.bits_per_ptr_block)
#define DATA_BLOCK_MASK		(cmsfs.data_block_mask)
#define BLKSIZE			(cmsfs.blksize)

#define READDIR_FILE_ENTRY	-1
#define READDIR_END_OF_DIR	-2
#define READDIR_DIR_ENTRY	-3
#define READDIR_MAP_ENTRY	-4

#define NULL_BLOCK		0
#define VAR_FILE_END		-1

#define LINEFEED_REQUIRED	((struct record *)-1)

#define RSS_HEADER_STARTED	0x1
#define RSS_HEADER_COMPLETE	0x2
#define RSS_DATA_BLOCK_STARTED	0x4
#define RSS_DATA_BLOCK_EXT	0x8

struct record_ext {
	/* start addr of the extension */
	off_t disk_start;
	/* length of extension in this disk block */
	int len;

	struct record_ext *prev;
	struct record_ext *next;
};

struct record {
	/* length of the complete record */
	int total_len;
	/* offset off first record block on the disk */
	off_t disk_start;
	/* bytes in first record block */
	int first_block_len;
	/* logical offset, dependent on line feed mode */
	off_t file_start;
	/* spanned record extension */
	struct record_ext *ext;
};

/* file object for operations that follow open */
struct file {
	/* pointer to the fst entry */
	struct fst_entry *fst;
	/* translate mode enabled */
	int translate;
	/* linefeed mode enabled */
	int linefeed;

	/* table of records */
	struct record *rec_table;

	int record_scan_state;

	int next_record_hint;
};

#define WALK_FLAG_LOOKUP	0x1
#define WALK_FLAG_READDIR	0x2

struct walk_file {
	int flag;
	char *name;
	char *type;
	void *buf;
	off_t addr;
	fuse_fill_dir_t filler;
};

#ifdef DEBUG_ENABLED
static void dump_record(struct record *rec, int nr)
{
	struct record_ext *rext = rec->ext;

	DEBUG("|\n");
	DEBUG("+- record [%d] file: %lx  total: %u",
		nr, rec->file_start, rec->total_len);
	DEBUG("  disk: %lx  first_len: %u\n",
		rec->disk_start, rec->first_block_len);
	if (rext != NULL) {
		do {
			DEBUG("|-+- extension [%p] disk: %lx  len: %u\n",
				rext, rext->disk_start, rext->len);
			rext = rext->next;
		} while (rext != NULL);
	}
}
#endif

static inline int is_directory(const char *name,
			       const char *type)
{
	if ((*(unsigned long long *) name) != FST_ENTRY_DIR_NAME)
		return 0;
	if ((*(unsigned long long *) type) != FST_ENTRY_DIR_TYPE)
		return 0;
	return 1;
}

static inline int is_allocmap(const char *name,
			      const char *type)
{
	if ((*(unsigned long long *) name) != FST_ENTRY_ALLOC_NAME)
		return 0;
	if ((*(unsigned long long *) type) != FST_ENTRY_ALLOC_TYPE)
		return 0;
	return 1;
}

static inline int is_file(const char *name,
			  const char *type)
{
	if ((*(unsigned long long *) name) == 0)
		return 0;
	/* Assumption: type=0 is not legal */
	if ((*(unsigned long long *) type) == 0)
		return 0;
	return 1;
}

static unsigned long dec_to_hex(unsigned long addr)
{
	register unsigned long res asm("2") = 0;
	register unsigned long _addr asm("4") = addr;

	asm volatile(
	"       cvb %0,0(0,%1)"
		: "=d" (res) : "d" (_addr) : "memory");
	return res & 0xffffffff;
}

static inline struct file *get_fobj(struct fuse_file_info *fi)
{
	return (struct file *) fi->fh;
}

/*
 * Set file position and read. Die on errors. Complain if the read crosses
 * block boundaries.
 */
static void __read(void *buf, size_t size, off_t addr)
{
	if (((addr + (off_t) size - 1) & ~DATA_BLOCK_MASK) >
	    (addr & ~DATA_BLOCK_MASK))
		DIE("__read: crossing blocks addr: %lx  size: %ld\n",
			addr, size);
	if ((size_t)addr < cmsfs.fdir)
		DIE("__read: invalid addr: %lx\n", addr);
	if ((size_t)addr > cmsfs.size)
		DIE("__read: invalid addr: %lx\n", addr);

	memcpy(buf, cmsfs.map + addr, size);
	return;
}

/*
 * Convert EDF date to time_t
 */
static time_t fst_date_to_time_t(char *date, int century)
{
	struct tm tm;
	time_t time;
	unsigned int res[6];
	unsigned long long num;
	int i;

	/*
	 * date : YY MM DD HH MM SS (decimal!)
	 * century: 0=19, 1=20, dead=21
	 * convert decimal to hex
	 */
	for (i = 0; i < 6; i++) {
		num = date[i];
		num <<= 4;
		num += 0xc;	/* plus */
		res[i] = dec_to_hex((unsigned long) &num);
	}

	memset(&tm, 0, sizeof(tm));
	tm.tm_year = res[0];
	tm.tm_mon = res[1];
	tm.tm_mday = res[2];
	tm.tm_hour = res[3];
	tm.tm_min = res[4];
	tm.tm_sec = res[5];

	/* prepare for mktime */
	tm.tm_hour--;
	tm.tm_mon--;
	if (century == FST_FLAG_CENTURY)
		tm.tm_year += 100;

	time = mktime(&tm);
	if (time == -1) {
		fprintf(stderr, COMP "mktime failed!\n");
		memset(&time, 0, sizeof(time));
	}
	return time;
}

/*
 * Read one FST entry into *fst from offset on disk addr and detect type
 *
 * Return values:
 * ret > 0 : disk address of additional FOP block
 * ret = -1 : file entry filled
 * ret = -2 : end of directory
 * ret = -3 : directory entry
 * ret = -4 : allocmap entry
 */
static int readdir(struct fst_entry *fst, off_t addr)
{
	BUG(addr & (sizeof(struct fst_entry) - 1));

	__read(fst, sizeof(*fst), addr);

	if (is_directory(fst->name, fst->type)) {
		/* check for multi-block directory */
		if (REAL(fst->fop) != addr)
			return REAL(fst->fop);
		return READDIR_DIR_ENTRY;
	}

	if (is_allocmap(fst->name, fst->type))
		return READDIR_MAP_ENTRY;

	if (is_file(fst->name, fst->type))
		return READDIR_FILE_ENTRY;

	return READDIR_END_OF_DIR;
}

/*
 * Return number of characters excluding trailing spaces
 */
static inline int strip_right(const char *str, int size)
{
	while (str[size - 1] == 0x20)
		size--;
	return size;
}

/*
 * Convert ASCII name to EBCDIC name
 */
static int encode_edf_name(const char *path, char *name, char *type)
{
	char *tmp;
	int dot_pos, tlen;

	/*
	 * path is ascii string "FILE.EXT"
	 * readdir return fst.name fst.type as EBCDIC including spaces pre-fill
	 * name and type with ascii spaces, remove dot and convert to EBCDIC.
	 */
	memset(name, 0x20, 8);
	memset(type, 0x20, 8);

	tmp = index(path, '.');
	/* filenames without a dot are invalid! */
	if (tmp == NULL)
		return -EINVAL;

	dot_pos = tmp - path;
	if (dot_pos == 0 || dot_pos > 8)
		return -EINVAL;
	memcpy(name, path, dot_pos);
	ebcdic_enc(name, name, 8);

	tlen = strlen(path) - (dot_pos + 1);
	if (tlen == 0 || tlen > 8)
		return -EINVAL;

	memcpy(type, path + dot_pos + 1, tlen);
	ebcdic_enc(type, type, 8);
	return 0;
}

/*
 * Convert EBCDIC name to ASCII name
 */
static void decode_edf_name(char *file, char *fname, char *ftype)
{
	int len_name, len_type;

	ebcdic_dec(fname, fname, 8);
	ebcdic_dec(ftype, ftype, 8);

	/* strip spaces but only from the end */
	len_name = strip_right(fname, 8);
	memcpy(file, fname, len_name);

	/* add dot 1 fluid */
	file[len_name] = '.';
	len_type = strip_right(ftype, 8);
	memcpy(&file[len_name + 1], ftype, len_type);
}

/*
 * Read pointer from fixed size pointer block and return
 * absolute address on disk.
 */
static off_t get_fixed_pointer(off_t addr)
{
	struct fixed_ptr ptr;

	if (!addr)
		return NULL_BLOCK;
	__read(&ptr, sizeof(ptr), addr);
	if (!ptr.next)
		return NULL_BLOCK;
	else
		return REAL((off_t)ptr.next);
}

/*
 * Read variable pointer from block and return absolute address on disk
 * and highest record number.
 */
static off_t get_var_pointer(off_t addr, int *max_record,
			     unsigned int *disp)
{
	struct var_ptr vptr;
	off_t ptr = 0;

	BUG(!addr);

	__read(&vptr, VPTR_SIZE, addr);
	ptr = (off_t) vptr.next;

	*max_record = vptr.max_record_nr;
	*disp = vptr.displacement;

	if (!ptr) {
		if (vptr.max_record_nr)
			return NULL_BLOCK;
		else
			return VAR_FILE_END;
	} else
		return REAL(ptr);
}

/*
 * Summarize the number of bytes used in the last data block.
 */
static int walk_last_var_data_block(off_t addr, size_t *total)
{
	uint16_t len;
	int left = cmsfs.blksize;

	/* subtract displacement */
	left -= addr & DATA_BLOCK_MASK;

	while ((size_t) left >= sizeof(len)) {

		__read(&len, sizeof(len), addr);

		/*
		 * Null length means no more records follow.
		 * Assumption: the last block is zero-padded.
		 */
		if (!len)
			return 0;

		/* add length of record with the header length */
		*total += len + sizeof(len);

		left -= len + sizeof(len);

		/* point to next record */
		addr += len + sizeof(len);
	}
	return 0;
}

static int skip_header_byte(struct file *f)
{
	if (f->fst->record_format == RECORD_LEN_FIXED)
		return 0;

	if (f->record_scan_state == RSS_HEADER_STARTED)
		return 1;
	else
		return 0;
}

static void set_record_len_upper(struct file *f, int record, uint8_t len)
{
	struct record *r = &f->rec_table[record];

	if (f->record_scan_state != RSS_DATA_BLOCK_STARTED &&
	    f->record_scan_state != RSS_DATA_BLOCK_EXT)
		DIE("set_record_len_upper\n");

	r->total_len = len << 8;
	f->record_scan_state = RSS_HEADER_STARTED;
}

static void set_record_len_lower(struct file *f, int record, uint8_t len)
{
	struct record *r = &f->rec_table[record];

	if (f->record_scan_state != RSS_HEADER_STARTED)
		DIE("set_record_len_lower\n");

	r->total_len += len;
	f->record_scan_state = RSS_HEADER_COMPLETE;
}

static void set_record_len(struct file *f, int record, uint16_t len)
{
	struct record *r = &f->rec_table[record];

	if (f->record_scan_state != RSS_DATA_BLOCK_STARTED &&
	    f->record_scan_state != RSS_DATA_BLOCK_EXT)
		DIE("set_record_len\n");

	r->total_len = len;
	f->record_scan_state = RSS_HEADER_COMPLETE;
}

static void set_record(struct file *f, int *record, off_t addr, int len,
		       size_t *total)
{
	struct record *r = &f->rec_table[*record];

	if (f->record_scan_state != RSS_HEADER_COMPLETE)
		DIE("set_record\n");

	DEBUG("record: %d\t (%p) at: %lx\t  len: %d\t", *record, r, addr, len);
	r->first_block_len = len;
	r->disk_start = addr;

	/* add previous record linefeed but not for the first record */
	if (f->linefeed && *record)
		(*total)++;
	r->file_start = *total;
	(*total) += r->total_len;
	f->record_scan_state = RSS_DATA_BLOCK_STARTED;
	DEBUG("  file offset: %lu\n", r->file_start);
}

static void add_record_ext(struct record *rec, struct record_ext *ext)
{
	struct record_ext *tmp;
	int i = 0;

	if (rec->ext == NULL) {
		DEBUG("add: extension 0 to record: %p\n", rec);
		rec->ext = ext;
		ext->prev = NULL;
		ext->next = NULL;
	} else {
		tmp = rec->ext;
		i++;
		while (tmp->next != NULL) {
			i++;
			tmp = tmp->next;
		}
		DEBUG("add: extension %d to record: %p\n", i, rec);
		tmp->next = ext;
		ext->prev = tmp;
		ext->next = NULL;
	}
}

static void set_record_extension(struct file *f, int *record, off_t addr,
				 int len)
{
	struct record_ext *ext;
	struct record *rec = &f->rec_table[*record];

	if (f->record_scan_state != RSS_DATA_BLOCK_STARTED &&
	    f->record_scan_state != RSS_DATA_BLOCK_EXT)
		DIE("set_record_extension\n");

	ext = malloc(sizeof(struct record_ext));
	if (ext == NULL)
		DIE_PERROR("malloc failed\n");
	ext->len = len - skip_header_byte(f);
	ext->disk_start = addr + skip_header_byte(f);

	DEBUG("record: extension block %lx to record %d len: %d\n",
		ext->disk_start, *record, ext->len);

	add_record_ext(rec, ext);
	f->record_scan_state = RSS_DATA_BLOCK_EXT;
}

static int end_of_file(struct file *f, int record)
{
	if (record == f->fst->nr_records)
		return 1;
	return 0;
}

static void walk_fixed_data_block(struct file *f, off_t addr,
				  int *record, size_t *total, int block)
{
	int left = cmsfs.blksize;
	off_t offset = block * cmsfs.blksize;
	int rlen = (f->fst->record_len > cmsfs.blksize) ?
		cmsfs.blksize : f->fst->record_len;
	int first = (offset % f->fst->record_len) ?
		f->fst->record_len - (offset % f->fst->record_len) : 0;

	DEBUG("walkf: rlen: %d  first: %d  offset: %lx\n", rlen, first, offset);
	BUG(addr & DATA_BLOCK_MASK);

	if (first) {
		set_record_extension(f, record, addr, first);
		left -= first;
		if (addr != NULL_BLOCK)
			addr += first;
	}

	while (left >= rlen) {
		/*
		 * Increment record number only after adding a possible
		 * extension. *record starts with -1 so the first is 0.
		 */
		(*record)++;
		if (end_of_file(f, *record))
			return;

		set_record_len(f, *record, f->fst->record_len);
		set_record(f, record, addr, rlen, total);
		f->record_scan_state = RSS_DATA_BLOCK_STARTED;

		left -= rlen;
		if (addr != NULL_BLOCK)
			addr += rlen;
	}

	/* partial record left */
	if (left) {
		(*record)++;
		if (end_of_file(f, *record))
			return;

		set_record_len(f, *record, f->fst->record_len);
		set_record(f, record, addr, left, total);
		f->record_scan_state = RSS_DATA_BLOCK_STARTED;
		return;
	}
}

static void walk_var_data_block(struct file *f, off_t addr, unsigned int disp,
				int *record, size_t *total)
{
	uint8_t half_len;
	uint16_t len;
	int left = cmsfs.blksize;

	BUG(addr & DATA_BLOCK_MASK);

	/*
	 * disp set means 1 or 2 header bytes and possibly data bytes on the
	 * last block or a null block
	 */
	if (disp) {
		if (addr == NULL_BLOCK) {
			/*
			 * Special case: record header on last block wo. data.
			 * That means no record data yet for this block.
			 */
			if (f->record_scan_state == RSS_HEADER_COMPLETE)
				set_record(f, record, addr, BLKSIZE, total);
			else
				set_record_extension(f, record, addr, BLKSIZE);
			return;
		}

		if (disp == VAR_RECORD_SPANNED)
			len = cmsfs.blksize;
		else
			len = disp;

		/* split header -> read second length byte */
		if (f->record_scan_state == RSS_HEADER_STARTED) {
			__read(&half_len, sizeof(half_len), addr);
			set_record_len_lower(f, *record, half_len);
			left--;
			addr++;
			len--;
		}

		if (f->record_scan_state == RSS_HEADER_COMPLETE)
			set_record(f, record, addr, len, total);
		else
			set_record_extension(f, record, addr, len);

		if (disp == VAR_RECORD_SPANNED)
			return;

		left -= len;
		addr += len;
	}

	/* at least one data byte */
	while ((size_t)left >= sizeof(len) + 1) {

		__read(&len, sizeof(len), addr);

		/*
		 * Null length means no more records follow.
		 * Assumption: the last block is zero-padded.
		 */
		if (!len)
			return;

		/*
		 * Increment record number only after adding a possible
		 * extension. *record starts with -1 so the first is 0.
		 */
		(*record)++;
		set_record_len(f, *record, len);

		/* account consumed header bytes */
		left -= sizeof(len);
		addr += sizeof(len);

		/* limit to block end */
		if (len > left)
			len = left;

		/* add length of record including the header length */
		set_record(f, record, addr, len, total);
		f->record_scan_state = RSS_DATA_BLOCK_STARTED;

		left -= len;
		/* point to next record header */
		addr += len;
	}

	/* 2 header bytes left */
	if (left == 2) {
		__read(&len, sizeof(len), addr);
		if (!len)
			return;

		(*record)++;
		set_record_len(f, *record, len);
		return;
	}

	/* split header */
	if (left == 1) {
		__read(&half_len, sizeof(half_len), addr);
		(*record)++;
		set_record_len_upper(f, *record, half_len);
		f->record_scan_state = RSS_HEADER_STARTED;
	}
}

static void cache_fixed_data_block(struct file *f, off_t addr, int *block,
				   int *record, size_t *total)
{
	/*
	 * Cannot distinguish null block pointers from not existing pointers,
	 * so this fn is called for the whole pointer block and maybe for
	 * non-existing blocks and records too. Check and bail out if EOF.
	 */
	if (*block >= f->fst->nr_blocks)
		return;

	DEBUG("cachef: caching data block %d  addr: %lx  record: %d\n",
		*block, addr, *record);

	walk_fixed_data_block(f, addr, record, total, *block);
	(*block)++;
}

/*
 * Walk all pointer blocks of a fixed file and call function for every
 * data block respecting the sequence of the data.
 */
static void cache_fixed_file(struct file *f, off_t addr, int level,
			     int *block, int *record, size_t *total)
{
	int left = PTRS_PER_BLOCK;
	off_t ptr;

	DEBUG("cachef: addr: %lx  level: %d  block: %d\n",
		addr, level, *block);
	if (level > 0) {
		level--;
		while (left--) {
			ptr = get_fixed_pointer(addr);
			cache_fixed_file(f, ptr, level, block, record, total);
			/* don't increment for null block pointers */
			if (addr)
				addr += PTR_SIZE;
		}
		return;
	}
	cache_fixed_data_block(f, addr, block, record, total);
}

static void cache_variable_data_block(struct file *f, off_t addr, int *block,
				      int *record, int disp, size_t *total)
{
	DEBUG("cachev: caching data block %d   addr: %lx  disp: %d\n",
		*block, addr, disp);
	(*block)++;
	walk_var_data_block(f, addr, disp, record, total);
}

/*
 * Walk all pointer blocks of a variable file and call function for every
 * data block respecting the sequence of the data.
 */
static void cache_variable_file(struct file *f, off_t addr, int level,
				int *block, unsigned int *disp,
				int *record, size_t *total)
{
	int nr, left = VPTRS_PER_BLOCK;
	off_t ptr;

	DEBUG("cachev: addr: %lx  level: %d\n", addr, level);

	if (level > 0) {
		level--;
		/* 4 or 8 bytes are left at the end (offset) which we ignore */
		while (left--) {
			ptr = get_var_pointer(addr, &nr, disp);
			if (ptr == VAR_FILE_END)
				return;
			cache_variable_file(f, ptr, level, block,
					    disp, record, total);
			addr += VPTR_SIZE;
		}
		return;
	}
	cache_variable_data_block(f, addr, block, record, *disp, total);
}

static void locate_last_data_vptr(off_t addr, int level,
				  struct fst_entry *fst, struct var_ptr *vptr)
{
	unsigned int last;

	if (!level)
		return;
	level--;

	/* read offset pointer from the end of the var pointer block */
	__read(&last, sizeof(last), addr + cmsfs.blksize - sizeof(last));

	BUG(last % VPTR_SIZE || last > cmsfs.blksize);

	__read(vptr, VPTR_SIZE, addr + last);

	BUG(vptr->max_record_nr != fst->nr_records);

	if (level == 0)
		/* vptr should contain the highest data block pointer */
		return;

	if (vptr->next == 0) {
		DEBUG("locate: found NULL pointer level: %d  addr: %lx\n",
			level + 1, addr);
		return;
	}

	locate_last_data_vptr(REAL(vptr->next), level, fst, vptr);
}

static int is_textfile(struct fst_entry *fst)
{
	unsigned int i;
	char type[9];

	if (!fst)
		return 0;

	memset(type, 0, 9);
	ebcdic_dec(fst->type, type, 8);

	for (i = 0; i < ARRAY_SIZE(conv_filetypes); i++)
		if (strncmp(type, conv_filetypes[i],
			    strlen(conv_filetypes[i])) == 0)
			return 1;
	return 0;
}

/* decide if linefeeds are needed for this file type */
static int linefeed_mode(struct fst_entry *fst)
{
	if (cmsfs.raw_mode)
		return 0;
	if (cmsfs.ascii_mode)
		return 1;
	return is_textfile(fst);
}

static void cache_fst_addr(off_t addr, const char *file)
{
	struct fcache_entry *fce;
	ENTRY e, *eptr;

	e.key = strdup(file);

	if (hsearch_r(e, FIND, &eptr, cmsfs.htab) == 0) {
		/* cache it */
		fce = &cmsfs.fcache[cmsfs.fcache_used];
		cmsfs.fcache_used++;

		fce->fst_addr = addr;

		e.data = fce;
		if (hsearch_r(e, ENTER, &eptr, cmsfs.htab) == 0)
			DIE("hsearch: hash table full\n");
	} else
		free(e.key);
}

/*
 * for each FST entry in a directory block...
 * Return:
 *	hit == NULL : lookup file not found
 *	hit != NULL : lookup file found, addr of the fst entry
 */
static void walk_dir_block(struct fst_entry *fst, struct walk_file *walk,
			   int level, off_t *hit)
{
	int ret, left;
	off_t ptr, addr = walk->addr;
	char file[18];	/* name + type + 2 zeros */

	DEBUG("dir: walk dir block at addr: %lx level: %d\n",
		walk->addr, level);

	/* handle higher level directory pointer blocks */
	if (level > 0) {
		level--;
		left = PTRS_PER_BLOCK;
		while (left--) {
			ptr = get_fixed_pointer(addr);
			if (!ptr)
				break;
			walk->addr = ptr;
			walk_dir_block(fst, walk, level, hit);
			if (hit != NULL && *hit)
				return;
			addr += PTR_SIZE;
		}
		return;
	}

	left = cmsfs.blksize / sizeof(struct fst_entry);
	while (left--) {
		ret = readdir(fst, walk->addr);

		/* directory and allocmap type are skipped */

		if (ret == READDIR_FILE_ENTRY) {
			if (walk->flag == WALK_FLAG_LOOKUP) {
				if ((strncmp(fst->name, walk->name, 8) == 0) &&
				    (strncmp(fst->type, walk->type, 8) == 0)) {
					/* got it */
					*hit = walk->addr;
					return;
				}
			}

			if (walk->flag == WALK_FLAG_READDIR) {
				bzero(file, sizeof(file));
				decode_edf_name(file, fst->name, fst->type);
				cache_fst_addr(walk->addr, file);
				walk->filler(walk->buf, file, NULL, 0);
			}
		}

		if (ret == READDIR_END_OF_DIR)
			break;
		walk->addr += sizeof(struct fst_entry);
	};
	return;
}

static void walk_directory(struct fst_entry *fst, struct walk_file *walk,
			   off_t *hit)
{
	if (cmsfs.levels == 0)
		walk->addr = cmsfs.fdir;
	else
		walk->addr = cmsfs.fop;
	walk_dir_block(fst, walk, cmsfs.levels, hit);
}

/*
 * Locate the file's fst_entry in any of the directory blocks
 */
static off_t lookup_file(const char *path, struct fst_entry *fst)
{
	char name[8], type[8];
	struct walk_file walk;
	struct fcache_entry *fce;
	ENTRY e, *eptr;
	off_t faddr = 0;

	e.key = strdup(path);

	/* already cached ? */
	if (hsearch_r(e, FIND, &eptr, cmsfs.htab)) {
		fce = eptr->data;
		/* read in the fst entry */
		__read(fst, sizeof(*fst), fce->fst_addr);
		free(e.key);
		return fce->fst_addr;
	}

	free(e.key);
	if (encode_edf_name(path, name, type))
		return 0;
	memset(&walk, 0, sizeof(walk));
	walk.flag = WALK_FLAG_LOOKUP;
	walk.name = name;
	walk.type = type;
	walk_directory(fst, &walk, &faddr);
	if (!faddr)
		return 0;
	cache_fst_addr(faddr, path);
	return faddr;
}

static size_t get_file_size_fixed(struct fst_entry *fst)
{
	return fst->nr_records * fst->record_len;
}

static size_t get_file_size_variable(struct fst_entry *fst)
{
	size_t total = 0;
	struct var_ptr vptr;
	off_t ptr;

	if (fst->levels > 0) {
		locate_last_data_vptr(REAL(fst->fop), fst->levels,
				      fst, &vptr);
		if (vptr.next == 0) {
			/* last block is a null block */
			total += cmsfs.blksize;
			goto skip;
		}
		ptr = REAL(vptr.next);
		if (vptr.displacement != VAR_RECORD_SPANNED) {
			ptr += vptr.displacement;
			/* count displacement as used space */
			total += vptr.displacement;
		}
	} else
		ptr = REAL(fst->fop);

	/* now count the remaining used space in the last block */
	walk_last_var_data_block(ptr, &total);

skip:
	/*
	 * Add the full blocks. For variable record file nr_blocks contains
	 * also null blocks.
	 */
	if (fst->nr_blocks)
		total += (fst->nr_blocks - 1) * cmsfs.blksize;
	return total;
}

/*
 * Return the file size as it is on the disk. Includes headers for
 * variable records.
 */
static size_t get_file_size(struct fst_entry *fst)
{
	if (fst->record_format == RECORD_LEN_FIXED)
		return get_file_size_fixed(fst);
	if (fst->record_format == RECORD_LEN_VARIABLE)
		return get_file_size_variable(fst);
	DIE("Invalid file format\n");
}

static size_t get_file_size_logical(struct fst_entry *fst)
{
	size_t total = get_file_size(fst);

	/* subtract the record headers */
	if (fst->record_format == RECORD_LEN_VARIABLE)
		total -= fst->nr_records * VAR_RECORD_HEADER_SIZE;

	if (linefeed_mode(fst))
		total += fst->nr_records;
	return total;
}

static int cmsfs_getattr(const char *path, struct stat *stbuf)
{
	struct fst_entry fst;
	int mask = 0440;
	off_t addr;

	DEBUG("cmsfs_getattr: path: %s\n", path);

	/* sanity check, if we can't read we can't do much */
	if (access(cmsfs.device, R_OK) != 0)
		return -EPERM;

	memset(stbuf, 0, sizeof(*stbuf));
	stbuf->st_uid = getuid();
	stbuf->st_gid = getgid();

	if (strcmp(path, "/") == 0) {
		stbuf->st_mode = S_IFDIR | mask | 0110;
		stbuf->st_nlink = 2;

		readdir(&fst, cmsfs.fdir);

		/* date */
		stbuf->st_mtime = fst_date_to_time_t(&fst.date[0],
			fst.flag & FST_FLAG_CENTURY);
		stbuf->st_atime = stbuf->st_ctime = stbuf->st_mtime;

		/* size */
		stbuf->st_size = fst.record_len * fst.nr_records;
	} else {
		/* path is always absolute */
		addr = lookup_file(path + 1, &fst);
		if (!addr)
			return -ENOENT;

		stbuf->st_mtime = stbuf->st_atime = stbuf->st_ctime =
			fst_date_to_time_t(&fst.date[0],
					   fst.flag & FST_FLAG_CENTURY);
		stbuf->st_size = get_file_size_logical(&fst);
		stbuf->st_mode = S_IFREG | mask;
		stbuf->st_nlink = 1;
	}
	return 0;
}

static int cmsfs_readdir(const char *path, void *buf, fuse_fill_dir_t filler,
			 off_t offset, struct fuse_file_info *fi)
{
	struct walk_file walk;
	struct fst_entry fst;

	(void) offset;
	(void) fi;

	/*
	 * Offset is ignored and 0 passed to the filler fn so the whole
	 * directory is read at once.
	 */

	/* EDF knows only the root directory */
	if (strcmp(path, "/") != 0)
		return -ENOENT;

	filler(buf, ".", NULL, 0);
	filler(buf, "..", NULL, 0);

	memset(&walk, 0, sizeof(walk));
	/* readdir is possible without open so fi->fh is not set */
	walk.flag = WALK_FLAG_READDIR;
	walk.buf = buf;
	walk.filler = filler;
	walk_directory(&fst, &walk, NULL);
	return 0;
}

static void cache_file(struct file *f)
{
	int block = 0;
	int record = -1;
	unsigned disp = 0;
	size_t total = 0;

	if (f->fst->record_format == RECORD_LEN_FIXED)
		cache_fixed_file(f, REAL(f->fst->fop), f->fst->levels, &block,
				 &record, &total);
	if (f->fst->record_format == RECORD_LEN_VARIABLE)
		cache_variable_file(f, REAL(f->fst->fop), f->fst->levels,
				    &block, &disp, &record, &total);
}

/*
 * Nice caveat: for fixed files nr_blocks is excluding null blocks,
 * for variable files nr_blocks is including null blocks.
 * Add null blocks for fixed files so allocation and file end
 * checks work identical for both variants.
 */
static void workaround_nr_blocks(struct file *f)
{
	int nr;

	if (f->fst->record_format == RECORD_LEN_VARIABLE)
		return;
	DEBUG("workaround: correcting nr_blocks from: %d", f->fst->nr_blocks);
	nr = f->fst->nr_records * f->fst->record_len / BLKSIZE;
	if (f->fst->nr_records * f->fst->record_len % BLKSIZE)
		nr++;
	f->fst->nr_blocks = nr;
	DEBUG(" -> to: %d\n", nr);
}

/*
 * Create a file object to cache all needed file data.
 * Note: the caller must ensure that the file exists.
 */
static struct file *create_file_object(struct fst_entry *fst, int *rc)
{
	struct file *f;

	f = malloc(sizeof(*f));
	if (f == NULL)
		goto oom;
	memset(f, 0, sizeof(*f));
	DEBUG("cache: file object at addr: %p-%lx\n",
		f, (unsigned long)f + sizeof(*f));

	f->fst = malloc(sizeof(struct fst_entry));
	if (f->fst == NULL)
		goto oom_f;
	DEBUG("cache: fst at addr: %p-%lx\n",
		f, (unsigned long)f->fst + sizeof(struct fst_entry));

	memcpy(f->fst, fst, sizeof(struct fst_entry));
	workaround_nr_blocks(f);

	f->linefeed = linefeed_mode(f->fst);
	f->translate = f->linefeed;
	f->record_scan_state = RSS_DATA_BLOCK_STARTED;

	f->rec_table = calloc(fst->nr_records, sizeof(struct record));
	if (f->rec_table == NULL)
		goto oom_fst;
	DEBUG("cache: record table at addr: %p\n", f->rec_table);

	cache_file(f);
	return f;

oom_fst:
	free(f->fst);
oom_f:
	free(f);
oom:
	*rc = -ENOMEM;
	return NULL;
}

static void destroy_file_object(struct file *f)
{
	struct record *rec;
	struct record_ext *rext, *tmp;
	int i;

	for (i = 0; i < f->fst->nr_records; i++) {
		rec = &f->rec_table[i];

		if (rec->ext == NULL)
			break;

		rext = rec->ext;
		do {
			tmp = rext->next;
			free(rext);
			rext = tmp;
		} while (rext != NULL);
	}

	free(f->rec_table);
	free(f->fst);
	free(f);
}

static int cmsfs_open(const char *path, struct fuse_file_info *fi)
{
	struct fst_entry fst;
	struct file *f;
	int rc;

	/*
	 * open flags:
	 * O_DIRECTORY: FUSE captures open on / so not needed.
	 * O_NOATIME ignored because there is no atime in EDF.
	 * O_NOFOLLOW can be ignored since EDF has no links.
	 * O_SYNC ignored for now since IO is alwasy sync.
	 */

	/* prevent usage of unsupported flags */
	if (fi->flags &	O_APPEND)
		return -ENOTSUP;

	/* path is always absolute */
	if (!lookup_file(path + 1, &fst))
		return -ENOENT;

	f = create_file_object(&fst, &rc);
	if (f == NULL)
		return rc;

	fi->fh = (uint64_t) f;
	return 0;
}

static int cmsfs_release(const char *path, struct fuse_file_info *fi)
{
	struct file *f = get_fobj(fi);

	(void) path;

	if (f == NULL) {
		DEBUG("release internal error\n");
		return -EINVAL;
	}

	destroy_file_object(f);
	fi->fh = 0;
	return 0;
}

/* guess position in record table */
static int guess_record_number(struct file *f, off_t offset)
{
	int nr;

	if (f->linefeed)
		nr = (offset / (f->fst->record_len + 1));
	else
		nr = (offset / f->fst->record_len);
	if (nr >= f->fst->nr_records)
		nr = f->fst->nr_records - 1;
	return nr;
}

static int offset_linefeed(off_t offset, struct record *prev,
			      struct record *next)
{
	if ((offset < next->file_start) &&
	    (offset >= prev->file_start + prev->total_len)) {
		DEBUG("find: found linefeed\n");
		return 1;
	}
	return 0;
}

static int offset_record(off_t offset, struct record *rec)
{
	if (offset >= rec->file_start &&
	    offset < rec->file_start + rec->total_len) {
		return 1;
	}
	return 0;
}

static void set_hint(struct file *f, int hint)
{
	f->next_record_hint = hint;

	/* limit hint to last record */
	if (f->next_record_hint >= f->fst->nr_records)
		f->next_record_hint = f->fst->nr_records - 1;
}

/*
 * Returns: record number in *nr
 *   > 0  : ptr to found record
 *   -1   : linefeed offset
 *   NULL : error
 */
static struct record *find_record(struct file *f, off_t offset, int *nr)
{
	struct record *rec;
	int i, start, step, max = f->fst->nr_records;

	/*
	 * next_record_hint is a guess which is optimal for sequential
	 * single-threaded reads.
	 */
	i = f->next_record_hint;
	rec = &f->rec_table[i];

	if (offset_record(offset, rec)) {
		DEBUG("find: next hint success: %d\n", i);
		/* increment hint for sequential read, fails for extensions */
		set_hint(f, f->next_record_hint + 1);
		*nr = i;
		return rec;
	}

	/* look out for previous record linefeed from sequential read hint */
	if (f->linefeed && i > 0)
		if (offset_linefeed(offset, &f->rec_table[i - 1], rec))
			return LINEFEED_REQUIRED;

	DEBUG("find: hint: %d failed for offset: %lx\n", i, offset);

	start = guess_record_number(f, offset);

	/* because of potential linefeed we need to check the next record */
	rec = &f->rec_table[start];
	if (offset < rec->file_start) {
		step = -1;
		DEBUG("find: switch direction start: %d  record start: %lx\n",
			start, rec->file_start);
	} else
		step = 1;

	for (i = start; i >= 0 && i < max; i += step) {
		rec = &f->rec_table[i];
		if (offset_record(offset, rec)) {
			DEBUG("find: found record: %d\n", i);
			set_hint(f, i + 1);
			*nr = i;
			return rec;
		}

		/* last record reached, only one linefeed can follow */
		if (i == max - 1) {
			if (f->linefeed &&
			    (offset == rec->file_start + rec->total_len))
				return LINEFEED_REQUIRED;
			else {
				DEBUG("find: wrong offset: %lu !\n", offset);
				return NULL;
			}
		}

		/* check for linefeed byte between two records */
		if (step == 1) {
			if (offset_linefeed(offset, rec, &f->rec_table[i + 1]))
				return LINEFEED_REQUIRED;
		} else
			/*
			 * No need to check if i > 0 since no linefeed before
			 * record 0 possible.
			 */
			if (offset_linefeed(offset, &f->rec_table[i - 1], rec))
				return LINEFEED_REQUIRED;

	}
	DEBUG("find: record not found!\n");
	return NULL;
}

/* get addr and block size from a record */
static void get_block_data_from_record(struct record *rec, off_t offset,
				       off_t *addr, int *chunk)
{
	struct record_ext *rext;
	int record_off = offset - rec->file_start;

	DEBUG("get: offset: %lu  file_start: %lu  file_end: %lu\n",
		offset, rec->file_start, rec->file_start + rec->total_len);
	if (record_off >= rec->first_block_len) {
		record_off -= rec->first_block_len;
		rext = rec->ext;

		BUG(rext == NULL);
		DEBUG("get: record offset in extension: record_off: %d\n",
			record_off);
		DEBUG("get: from record %p  ext disk_start: %lx  len: %d\n",
			rec, rext->disk_start, rext->len);

		while (record_off >= rext->len) {
			record_off -= rext->len;
			rext = rext->next;
			BUG(rext == NULL);
			DEBUG("get: from record %p  ext disk_start: %lx  len: %d\n",
				rec, rext->disk_start, rext->len);
		}
		DEBUG("get: found extension %p  ext disk_start: %lx  len: %d  record_off: %d\n",
			rext, rext->disk_start, rext->len, record_off);

		/* found the right record extension */
		if (rext->disk_start == NULL_BLOCK)
			*addr = NULL_BLOCK;
		else
			*addr = rext->disk_start + record_off;
		*chunk = rext->len - record_off;
	} else {
		if (rec->disk_start == NULL_BLOCK)
			*addr = NULL_BLOCK;
		else
			*addr = rec->disk_start + record_off;
		*chunk = rec->first_block_len - record_off;
	}
}

static int cmsfs_read(const char *path, char *buf, size_t size, off_t offset,
		      struct fuse_file_info *fi)
{
	size_t len, copied = 0;
	int chunk, nr;
	off_t addr;
	struct file *f = get_fobj(fi);
	struct record *rec;

	(void) path;

	DEBUG("cmsfs_read: path: %s  offset: %lx  size: %lu\n",
		path, offset, size);
	len = get_file_size_logical(f->fst);

	if ((size_t) offset >= len)
		return 0;

	if (offset + size > len)
		size = len - offset;
	DEBUG("read: set size to: %lu\n", size);

	while (size > 0 && copied < len) {
		rec = find_record(f, offset, &nr);
		if (rec == NULL) {
			copied = -EINVAL;
			DEBUG("read: invalid addr, size: %lu  copied: %lu  len: %lu\n",
				size, copied, len);
			goto out;
		}

		/* write linefeed directly to buffer and go to next record */
		if (rec == LINEFEED_REQUIRED) {
			BUG(!f->linefeed);
			*buf = '\n';
			buf++;
			copied++;
			offset++;
			size--;
			continue;
		}

#ifdef DEBUG_ENABLED
		dump_record(rec, nr);
#endif

		/* get addr and block size from record */
		get_block_data_from_record(rec, offset, &addr, &chunk);
		if (chunk <= 0 || addr < 0)
			DIE("get_block_data_from_record error\n");
		DEBUG("read: from: %lx  len: %u\n", addr, chunk);

		/* copy less if there is not enough space in the fuse buffer */
		if (size < (size_t) chunk)
			chunk = size;

		/* read one record */
		if (addr == NULL_BLOCK)
			memset(buf, 0, chunk);
		else
			__read(buf, chunk, addr);

		if (f->translate)
			ebcdic_dec(buf, buf, chunk);

		copied += chunk;
		size -= chunk;
		buf += chunk;
		offset += chunk;
	}
out:
	DEBUG("read copied: %lu\n", copied);
	return copied;
}

static int cmsfs_statfs(const char *path, struct statvfs *buf)
{
	unsigned int free_blocks = cmsfs.total_blocks - cmsfs.used_blocks;
	unsigned int inode_size = cmsfs.blksize + sizeof(struct fst_entry);

	(void) path;

	buf->f_bsize = buf->f_frsize = cmsfs.blksize;
	buf->f_blocks =	cmsfs.total_blocks;
	buf->f_bfree =  buf->f_bavail = free_blocks;
	/* number of possible inodes */
	buf->f_files = cmsfs.total_blocks * cmsfs.blksize / inode_size;

	buf->f_ffree = free_blocks * cmsfs.blksize / inode_size;
	buf->f_namemax = 17;
	return 0;
}

static struct fuse_operations cmsfs_oper = {
	.getattr	= cmsfs_getattr,
	.statfs		= cmsfs_statfs,
	.readdir	= cmsfs_readdir,
	.open		= cmsfs_open,
	.release	= cmsfs_release,
	.read		= cmsfs_read,
};

static int cmsfs_fuse_main(struct fuse_args *args,
			   struct fuse_operations *cmsfs_oper)
{
#if FUSE_VERSION >= 26
	return fuse_main(args->argc, args->argv, cmsfs_oper, NULL);
#else
	return fuse_main(args->argc, args->argv, cmsfs_oper);
#endif
}

static int cmsfs_process_args(void *data, const char *arg, int key,
			      struct fuse_args *outargs)
{
	(void)data;

	switch (key) {
	case FUSE_OPT_KEY_OPT:
		return 1;
	case FUSE_OPT_KEY_NONOPT:
		if (cmsfs.device == NULL) {
			cmsfs.device = strdup(arg);
			return 0;
		}
		return 1;
	case KEY_HELP:
		usage(outargs->argv[0]);
		fuse_opt_add_arg(outargs, "-ho");
		cmsfs_fuse_main(outargs, &cmsfs_oper);
		exit(0);
	case KEY_VERSION:
		fprintf(stderr, COMP "FUSE file system for CMS disks "
			"program version %s\n", RELEASE_STRING);
		fprintf(stderr, "Copyright IBM Corp. 2009\n");
		fuse_opt_add_arg(outargs, "--version");
		exit(0);

	default:
		DIE("Process arguments error\n");
	}
}

static int get_fop(off_t addr)
{
	struct fst_entry fst;

	__read(&fst, sizeof(fst), addr);
	return REAL(fst.fop);
}

static int get_levels(off_t addr)
{
	struct fst_entry fst;

	__read(&fst, sizeof(fst), addr);
	return fst.levels;
}

static int get_files_count(off_t addr)
{
	struct fst_entry fst;

	__read(&fst, sizeof(fst), addr);
	return fst.nr_records - 2; /* ignore director and allocmap entries */
}

static int get_order(int shift)
{
	int count = 0;
	while (!(shift & 0x1)) {
		shift >>= 1;
		count++;
	}
	return count;
}

static unsigned int get_allocmap(unsigned int fdir)
{
	unsigned int amap;

	__read(&amap, sizeof(amap),
		fdir + sizeof(struct fst_entry) + FST_FOP_OFFSET);
	return amap;
}

static void map_device(int fd)
{
	/* fstat on block device says st_size = 0... */
	lseek(fd, 0, SEEK_SET);
	cmsfs.size = lseek(fd, 0, SEEK_END);
	DEBUG("mmap size: %lu", cmsfs.size);

	/* map the whole block device for speeding-up read access */
	cmsfs.map = mmap(NULL, cmsfs.size, PROT_READ, MAP_SHARED, fd, 0);
	if (cmsfs.map == MAP_FAILED)
		DIE_PERROR("mmap failed");
	DEBUG("  addr: %p\n", cmsfs.map);
}

static void cmsfs_init(int fd)
{
	map_device(fd);

	/* calculate blocksize dependent values */
	cmsfs.data_block_mask = cmsfs.blksize - 1;

	cmsfs.fixed_ptrs_per_block = cmsfs.blksize / sizeof(int);
	cmsfs.var_ptrs_per_block = cmsfs.blksize / sizeof(struct var_ptr);

	cmsfs.bits_per_data_block = get_order(cmsfs.blksize);
	cmsfs.bits_per_ptr_block = get_order(cmsfs.fixed_ptrs_per_block);

	/* store directory information */
	cmsfs.levels = get_levels(cmsfs.fdir);
	cmsfs.fop = get_fop(cmsfs.fdir);
	cmsfs.files = get_files_count(cmsfs.fdir);
	DEBUG("cmsfs: files count for disk: %d\n", cmsfs.files);
	/* alloc cache entries for all files */
	cmsfs.fcache = calloc(cmsfs.files + 10, sizeof(struct fcache_entry));

	cmsfs.amap = REAL(get_allocmap(cmsfs.fdir));

	if (cmsfs.filetype_mode && cmsfs.ascii_mode)
		DIE("Incompatible options, choose either -t or -a\n");

	/* default mode is raw (keep EBCDIC) */
	if (!cmsfs.ascii_mode && !cmsfs.filetype_mode)
		cmsfs.raw_mode = 1;

	cmsfs.htab = malloc(sizeof(struct hsearch_data));
	if (cmsfs.htab == NULL)
		DIE_PERROR("malloc failed");
	memset(cmsfs.htab, 0, sizeof(struct hsearch_data));

	if (!hcreate_r(cmsfs.files + 10, cmsfs.htab))
		DIE("hcreate failed\n");
}

int main(int argc, char *argv[])
{
	struct fuse_args args = FUSE_ARGS_INIT(argc, argv);
	char *fsname;
	int rc, fd;

	if (fuse_opt_parse(&args, &cmsfs, cmsfs_opts,
			   cmsfs_process_args) == -1)
		DIE("Failed to parse option\n");

	if (!cmsfs.device)
		DIE("Missing device\n"
		    "Try '%s --help' for more information\n", argv[0]);

#ifdef DEBUG_ENABLED
	logfile = fopen("cmsfs-fuse.log", "w");
#endif

	DEBUG("using device: %s", cmsfs.device);
	fd = get_device_info(&cmsfs);
	DEBUG("  blocksize: %d\n", cmsfs.blksize);

	fsname = malloc(FSNAME_MAX_LEN);
	if (fsname == NULL)
		DIE_PERROR("malloc failed");

#if FUSE_VERSION >= 27
	snprintf(fsname, FSNAME_MAX_LEN, "-osubtype=cmsfs,fsname=%s",
		 cmsfs.device);
#else
	snprintf(fsname, FSNAME_MAX_LEN, "-ofsname=%s", cmsfs.device);
#endif
	fuse_opt_add_arg(&args, fsname);
	free(fsname);
	fuse_opt_add_arg(&args, "-oro");

	cmsfs_init(fd);
	cmsfs.fd = fd;

	rc = cmsfs_fuse_main(&args, &cmsfs_oper);

	fuse_opt_free_args(&args);
#ifdef DEBUG_ENABLED
	fclose(logfile);
#endif
	free(cmsfs.fcache);
	hdestroy_r(cmsfs.htab);
	free(cmsfs.htab);
	return rc;
}
